/* functions called by lm_auto, lm_lods for allocating and freeing arrays  

 *  allocz_y(), alloc_ibd(), alloc_az_prob(), alloc_ld_prob(),
 *  free_y(),   free_ibd(),  free_az_prob(),  free_ld_prob()
 *                                                              MAJ Aug 01  */
void  allocz_y (int ****pyy, int nl, int pdsz)
{
   int   j;

   *pyy = malloc (2 * sizeof (int **));
   alloc_chk (WHERE, *pyy);
   (*pyy)[0] = malloc ((nl + 1) * sizeof (int *));
   alloc_chk (WHERE, (*pyy)[0]);
   (*pyy)[1] = malloc ((nl + 1) * sizeof (int *));
   alloc_chk (WHERE, (*pyy)[1]);
   for (j = 0; j <= nl; ++j)
   {
      (*pyy)[0][j] = calloc (pdsz + 1, sizeof (int));
      alloc_chk (WHERE, (*pyy)[0][j]);
      (*pyy)[1][j] = calloc (pdsz + 1, sizeof (int));
      alloc_chk (WHERE, (*pyy)[1][j]);
   }
   return;
}
void  alloc_ibd (int **g, int ***ix, int ***lab, int ***lst, int **np,
                 double *****nsc)
{
/*  int g   [G]
 *  int ix  [G+1][CD+1]
 *  int lab [G+1][PT+1]        
 *  int lst [G+1][PT+1]
 *  int npt [G+1]
 *  double nsc [NC][NS][NL][PT+1]    ??? int ??? 
 *
 *  see table in limdefs.h for no. gametes/codes/patterns
 */

   int   i, j, k;

   *np = malloc ((MxGam + 1) * sizeof (int));
   alloc_chk (WHERE, *np);
   (*np)[0] = 0;

   *g = malloc (MxGam * sizeof (int));
   alloc_chk (WHERE, *g);

   *lab = malloc ((MxGam + 1) * sizeof (int *));
   alloc_chk (WHERE, *lab);
   (*lab)[0] = NULL;

   *ix = malloc ((MxGam + 1) * sizeof (int *));
   alloc_chk (WHERE, *ix);
   (*ix)[0] = NULL;

   *lst = malloc ((MxGam + 1) * sizeof (int *));
   alloc_chk (WHERE, *lst);
   (*lst)[0] = NULL;

   /* the index, list and label arrays are allocated as ragged arrays */

   for (i = 1; i <= MxGam; ++i)
   {
      (*ix)[i] = malloc ((CDgg[i] + 1) * sizeof (int));
      alloc_chk (WHERE, (*ix)[i]);
      (*lst)[i] = malloc ((PTgg[i] + 1) * sizeof (int));
      alloc_chk (WHERE, (*lst)[i]);
      (*lab)[i] = malloc ((PTgg[i] + 1) * sizeof (int));
      alloc_chk (WHERE, (*lab)[i]);
   }
   *nsc = malloc (n_compnts * sizeof (double ***));
   alloc_chk (WHERE, *nsc);

   /* the nscore array is ragged in 2 dimensions */

   for (i = 0; i < n_compnts; ++i)
   {
      (*nsc)[i] = malloc (NScSet[i] * sizeof (double **));
      alloc_chk (WHERE, (*nsc)[i]);
      for (j = 0; j < NScSet[i]; ++j)
      {
         (*nsc)[i][j] = malloc (NLoci * sizeof (double *));
         alloc_chk (WHERE, (*nsc)[i][j]);
         for (k = 0; k < NLoci; ++k)
         {
            (*nsc)[i][j][k] = malloc ((PTgg[NPrGam[i][j]] + 1) *
                                      sizeof (double));
            alloc_chk (WHERE, (*nsc)[i][j][k]);
         }
      }
   }
   return;
}
void  alloc_az_prob (int ***c, int ****rc, double ***pr1, double ***pr2,
                     double ***pr3, double ***pr4, double **s)
/* int c      [nc][nl]
 * int rc     [nc][nl][6]
 * double pr1 [nc][nl]
 * double pr2 [nc][nl]
 * double pr3 [nc][nl]
 * double pr4 [nc][nl]
 * double s   [nl]
 */
{

   ali_nc_nl (c);
   ali_nc_nl_n (rc, 6);
   ald_nc_nl (pr1);
   ald_nc_nl (pr2);
   ald_nc_nl (pr3);
   ald_nc_nl (pr4);
   ald_nl (s);

   return;
}
void  alloc_ld_prob (int **sc, int ****rc, int ***lsf, double ***lrat,
                     double ***clik, double ***pr, double ***pr1,
                     double ***pr2, double ***pr3, double ***pr4,
                     double ***pr5, double ***pr6)
{
   ali_nc (sc);
   ali_nc_nl_n (rc, 6);
   ali_n_nl (lsf, 3);

   ald_nc_n (lrat, 3);
   ald_nc_n (clik, 3);
   ald_nc_nl (pr);
   ald_nc_nl (pr1);
   ald_nc_nl (pr2);
   ald_nc_nl (pr3);
   ald_nc_nl (pr4);
   ald_nc_nl (pr5);
   ald_nc_nl (pr6);

   return;
}
void  free_y (int ****pyy, int nl)
{
   int   j;

   if (!*pyy)
      return;

   for (j = 0; j <= nl; j++)
   {
      free ((*pyy)[0][j]);
      free ((*pyy)[1][j]);
   }
   free ((*pyy)[0]);
   free ((*pyy)[1]);
   free (*pyy);
   *pyy = NULL;

   return;
}
void  free_ibd (int **g, int ***ix, int ***lab, int ***lst, int **np,
                double *****nsc)
{
   int   i, j, k;

   if (!*g)
      return;

   for (i = 0; i <= MxGam; ++i)
   {
      if (*ix && (*ix)[i])
         free ((*ix)[i]);
      if (*lst && (*lst)[i])
         free ((*lst)[i]);
      if (*lab && (*lab)[i])
         free ((*lab)[i]);
   }
   free (*ix);
   *ix = NULL;
   free (*lst);
   *lst = NULL;
   free (*lab);
   *lab = NULL;
   free (*np);
   *np = NULL;
   free (*g);
   *g = NULL;

   if (*nsc)
   {
      for (i = 0; i < n_compnts; ++i)
         if ((*nsc)[i])
         {
            for (j = 0; j < NScSet[i]; ++j)
               if ((*nsc)[i][j])
               {
                  for (k = 0; k < NLoci; ++k)
                     if ((*nsc)[i][j][k])
                        free ((*nsc)[i][j][k]);
                  free ((*nsc)[i][j]);
               }
            free ((*nsc)[i]);
         }
      free (*nsc);
      *nsc = NULL;
   }
   return;
}
void  free_az_prob (int ***c, int ****rc, double ***pr1, double ***pr2,
                    double ***pr3, double ***pr4, double **s)
{
   fre_nc_n ((void *) c);
   fre_nc_nl_n ((void *) rc);
   fre_nc_n ((void *) pr1);
   fre_nc_n ((void *) pr2);
   fre_nc_n ((void *) pr3);
   fre_nc_n ((void *) pr4);
   fre_n ((void *) s);
}
void  free_ld_prob (int **sc, int ****rc, int ***lsf, double ***lrat,
                    double ***clik, double ***pr, double ***pr1,
                    double ***pr2, double ***pr3, double ***pr4,
                    double ***pr5, double ***pr6)
{
   fre_n ((void *) sc);
   fre_nc_nl_n ((void *) rc);
   fre_n_n ((void *) lsf, 3);
   fre_nc_n ((void *) lrat);

   fre_nc_n ((void *) clik);
   fre_nc_n ((void *) pr);
   fre_nc_n ((void *) pr1);
   fre_nc_n ((void *) pr2);
   fre_nc_n ((void *) pr3);
   fre_nc_n ((void *) pr4);
   fre_nc_n ((void *) pr5);
   fre_nc_n ((void *) pr6);

   return;
}
