/*******************************************************************************
 program name: lm_lods.c
  
 Program to do lod scores (replaces homozyg.c from old Autozyg)
 The state (X) consists of segregation indicators -- so can generalize
    to cases of more than one observed/pedigree, and multiple loci.
  
 The data, Y, is the multilocus genotypes of observed individuals
    in the pedigree.

 Obtained by modification of jnt_auto.c                            EAT Dec 1998

 Here is the plan!
   OK this is the PLAN  (12/15/98)

 OK, now we have a general autozyg program, we need to go back to
 lod estimation, for those as want.  We use the same program (modified).

 For k markers, we will have (k+2) simulation points, and (k+2) evaluation
 points  (but maybe it is not worth evaluating all for all).
 To start we will just evaluate relative to a left-point and a right-point
 for each simulation point.

 The simulation points are mid-point of each interval, and r=0.1 to each
 end of chromosome (0.1 can be adjusted later), and "unlinked" which is
 0.5 to either end.

 Since the data are locus-by-locus, the likelihood ratio contribution is
 just ratio of P_theta(S)/P_sim_th(S).

 This is very easy, once we get the chromosome set-ups fixed.

 1) Trait locus data set-up -- assume trait-locus genotypes for observed
 individuals -- done  12/20/98
 2) Likelihood-ratio contributions routine,  -- done 12/18/98
 3) Simulation and evaluation chromosomes for each location. -- done 12/23/98

 4) Need a test run!! -- JV4C maybe.   working! 12/24/98

 New routines:  (12/19/98)
    full_seg_prob() and full_pen_seg_lod() and lik_rat()
       (all in full_logprob.c)
    full_seg_prob modified from old routine -- returns log(prob P(S)) 
       of all meioses on given component pedigree
    full_pen_seg_lod computes full penetrance P(Y|S) for data y, and
       given meiosis indicators, on given component pedigree.
          calls full_lod and returns full log-probability of (Y,S).
       [lod_next routine is not currently used ]

 Now to integrate the L-sampler, using the new Lsampler and Msampler
    libraries and new routines;                                    EAT Mar 2000

 Converted to MORGAN_V2.3                                          MAJ Jun 2000

 Modifications to specify simulation points, within each interval and
 outside of map.                                                   EAT Nov 2000 
 Then converted to MORGAN 2.3                                      MAJ Nov 2000

 Modification of full_seg_prob, and recombinant counts             EAT Nov 2000
    scoring of counts and tot_log_probs by component and locus/interval.

 Modifications to lik_rat, do_lsmpl, etc.                          EAT Dec 2000

 Modifications to use specify intervals for l-ratio updating and use
    more efficient seg_prob updating routines for l-sample steps   EAT Jan 2001

 Added changes for sequential imputation set-up                    AWG Mar 2001

 Converted fixed arrays to dynamically allocated (and freed) arrays
                                                                   MAJ Aug 2001
*******************************************************************************/

#define PNAM  "lm_lods"

#include "autozyg.h"
#include "set_auto.c"
#include "alloc_auto.c"

int   main (int argc, char *argv[])
{
   int ***y;                            /* full genotypic array */
   int   i, j, jj, k, l;
   int **dodo = NULL;
   double **quant = NULL;
   int   t, c, h, err;
   int   ngenes = 0, nfou, nloc;
   int  *nns, nsw, sx;
   int  *perm_sw, *perm_loc;            /* permutation arrays of switches 
                                         * and loci  */
   int  *scounts;                       /* [nc]         */
   int ***recounts;                     /* [nc][nl][6]  */
   int **lsflag;                        /* [3][nl]      flags for lsampler 
                                         *              l_ratio adjustments */
   double **l_ratio;                    /* [nc][3]      now this is the 
                                         *              LOG-l-ratio EAT 12/00 */
   double **prob;                       /* [nc][nl]    */
   double **cumulik;                    /* [nc][3]     */
   double **temp_log_prob_seg, **temp2_log_prob_seg;  /* [nc][nloc]  */
   double **tot_log_prob_seg, **tot_log_prob_pen;  /* [nc][nloc]  */
   double **tot_log_left_seg, **tot_log_rght_seg;  /* [nc][nloc]  */
   int   tot_count;

   double ***likeratios = NULL;         /* Matrix for likelihood ratios */
   double *likes = NULL;
   double **lods = NULL;                /* Lods scores */
   double leftratio = 0.0;
   double rightratio = 0.0;
   double like_eigenvalue = 0.0;

   double rec0[3] =
   {.5, .5, .5};
   int   loc0[3] =
   {0, 0, 0};

   FILE *pedF, *oSeedF;

   atexit (print_tail);
   PCleanUp = clean_auto;
   atexit (clean_morg);
   print_head (*argv);

   if (!proc_pars (argc, argv, PNAM))   /* process input parameter statements */
      exit (EXIT_FAILURE);

   if (!proc_auto ())                   /* check statements for completeness  */
      exit (EXIT_FAILURE);              /* completeness & consistency         *
                                         * setup controls for lm_lods         */
   if (!proc_seeds ())                  /* set up seed parameters             */
      exit (EXIT_FAILURE);

   if (!proc_fils ())                   /* set up pedigree file parameters    */
      exit (EXIT_FAILURE);

   setSeeds (SAMPL);                    /* initializes seeds for rans()       */

   if (!chk_pedfil ())                  /* checks pedigree file for errors    */
      give_up (ERR_qit);

   if (!mak_opedfil ())                 /* writes new pedigree file           */
      give_up (ERR_qit);                /* if there are changes               */

   if (!(pedF = open_ped ()))           /*opens pedigree file                 */
      give_up (ERR_qit);

   alloc_ped ();
   input_ped (pedF, 0, dodo, quant);    /*input pedigree and trait data       */

   nloc = NLoci - 1;

   nfou = count_founders ();
   printf (" Number of founders:  %d\n\n", nfou);

   MaxNsw = 2 * (pedsize - nfou);
   nns = malloc ((n_compnts + 1) * sizeof (int));
   alloc_chk (WHERE, nns);

   qdex = calloc (pedsize + 1, sizeof (int));
   alloc_chk (WHERE, qdex);
   trtphen = calloc (pedsize + 1, sizeof (int));
   alloc_chk (WHERE, qdex);

   /* now allocate space for genotypes y for our values of pedsize, nloc */

   allocz_y (&y, nloc, pedsize);

   /* now set marker data for observed individuals */

   set_markdat (y, nloc, n_compnts);

   /* now set trait data  */

   set_traitdat (y, qdex, trtphen, pedsize, n_compnts);

   allocate (pedsize, nloc);

   /* now set and index switches */

   nsw = list_switches (nloc, nns);
   printf (" Number of meiosis switches to sample is %d\n", nsw);
   fflush (stdout);

   /* set up permutation arrays for loci and switches */

   if (Scan)
   {
      perm_sw = malloc (nsw * sizeof (int));
      alloc_chk (WHERE, perm_sw);
      for (i = 0; i < nsw; ++i)
         perm_sw[i] = i;
      perm_loc = malloc ((nloc + 1) * sizeof (int));
      alloc_chk (WHERE, perm_loc);
      for (i = 0; i <= nloc; ++i)
         perm_loc[i] = i;
   }
   /* Use L-sampler or sequential imputation for set_up and for
      initial xx0, xx1 (maternal and paternal segregation indicators) */

   if (SmplSetup == LSMPL)              /* use L-sampler for set-up */
      set_lsmpl (0, nloc, NAll, Qq, y);
   else
   {                                    /* use sequential imputation */
      for (sx = FEM; sx <= MAL; sx++)
         set_chromosomes (nloc, Marek[sx], 2, loc0, rec0, 2 - sx);
      seq_imp (NSEQIMP, pedsize, nloc + 1, NAll, Qq, y, 0,
               (double *) 0, 1.0);      /* (AWG) */
   }

   /* Allocate memory for likeratios and lods scores */

   likes = (double *) malloc (NTPos * sizeof (double));
   alloc_chk (WHERE, likes);
   lods = (double **) malloc (n_compnts * sizeof (double *));
   alloc_chk (WHERE, lods);
   for (i = 0; i < n_compnts; ++i)
   {
      lods[i] = (double *) malloc ((NTPos - 1) * sizeof (double));
      alloc_chk (WHERE, lods[i]);
   }
   likeratios = (double ***) malloc (n_compnts * sizeof (double **));
   alloc_chk (WHERE, likeratios);
   for (i = 0; i < n_compnts; ++i)
   {
      likeratios[i] = (double **) malloc (NTPos * sizeof (double *));
      alloc_chk (WHERE, likeratios[i]);
      for (j = 0; j < NTPos; ++j)
      {
         likeratios[i][j] = (double *) malloc (NTPos * sizeof (double));
         alloc_chk (WHERE, likeratios[i][j]);
      }
   }
   alloc_ld_prob (&scounts, &recounts, &lsflag, &l_ratio, &cumulik, &prob,
                  &temp_log_prob_seg, &temp2_log_prob_seg, &tot_log_prob_seg,
                  &tot_log_prob_pen, &tot_log_left_seg, &tot_log_rght_seg);

   peel_trace = TRACE_LEVEL_0;          /* Turn on trace if level >0 */

   /* set genes from initial switches */

   for (j = k = 0; j <= nloc; ++j)
   {
      ngenes = set_genes (j);

      /* Allocate memory for gen_pen() */

      if (loci[j].locus_type == TRAIT_LOCUS)
         alloc_gen_pen (ngenes, j);
      else if (!k++)
         alloc_seg_pen (ngenes, nloc + 1, y);

      for (c = 0; c < n_compnts; ++c)
         prob[c][j] = seg_pen2 (c, y, Qq[j], j, ngenes, &err);
   }
   alloc_lpflag (nloc, 2);
   alloc_logprob ();
   alloc_msmpl ();

   /*                                            *** now start BIG loop *** */

   for (k = 0; k < NTPos; ++k)
   {
      printf ("\n\n CHROMOSOME SET %d\n\n", k);
      fflush (stdout);

      /* two evaluation sets for each simulation chromo (left/right) */

      printf (" Recombinations:\n");

      for (sx = FEM; sx <= MAL; ++sx)
      {
         loc0[0] = IntVal[k];
         rec0[0] = RecVal[sx][k];

         if (k == 0)
         {
            loc0[1] = nloc;
            rec0[1] = RecVal[sx][NTPos - 1];
         }
         else
         {
            loc0[1] = IntVal[k - 1];
            rec0[1] = RecVal[sx][k - 1];
         }

         loc0[2] = IntVal[k + 1];
         rec0[2] = RecVal[sx][k + 1];

         set_chromosomes (nloc, Marek[sx], 2, loc0, rec0, 2 - sx);

         set_lpflag (nloc, 2, 2 - sx);

         for (i = 0; i <= 2; ++i)
         {
            if (i == 0)
               printf ("    sim,   ");
            else if (i == 1)
               printf ("    left,  ");
            else
               printf ("    right, ");
            if (sx == MAL)
               printf ("M:");
            if (sx == FEM)
               printf ("F:");
            for (j = 0; j < nloc; ++j)
            {
               if (!(j % 6) && j)
                  printf ("\n             ");
               printf ("%3d%s", chromo[i][j],
                       dub2nlz (xecomo[2 - sx][i][j], 7, 3));
            }
            printf ("%3d\n", chromo[i][nloc]);
         }
      }
      printf ("\n");

      /* L-sampler set-up   */

      set_positions (nloc);             /* EAT  March 2000    */

      /* now should do an L-sample step on the trait locus (0) to reset  */

      do_lsmpl (nloc, 0, prob, y, ngenes);

      for (c = 0; c < n_compnts; ++c)
      {
         for (h = 0; h <= 2; ++h)
         {
            l_ratio[c][h] = 0.0;
            cumulik[c][h] = 0.0;
         }

         for (h = 0; h <= nloc; ++h)
         {
            tot_log_prob_seg[c][h] = 0.0;
            tot_log_left_seg[c][h] = 0.0;
            tot_log_rght_seg[c][h] = 0.0;
            tot_log_prob_pen[c][h] = 0.0;
         }
         scounts[c] = 0;
         for (j = 0; j <= nloc; ++j)
         {
            for (jj = 0; jj < 6; ++jj)
               recounts[c][j][jj] = 0;
         }
      }
      tot_count = 0;

      /* set initial l_ratio    */

      for (c = 0; c < n_compnts; ++c)
      {
         l_ratio[c][1] = 0.0;
         l_ratio[c][2] = l_ratio[c][1];

         for (h = 1; h <= 2; ++h)
         {
            for (j = 0; j < nloc; ++j)
               lsflag[h][j] = 1;
            l_ratio[c][h] +=
               diff_seg_prob (nloc, c, h, lsflag[h], temp_log_prob_seg);
         }
      }

      /* now do MCMC resampling by meiosis or by locus */

      for (t = 0; t < NIt; ++t)
      {
         if (rans () < LSamplProb)
         {                              /* ***    Do L-sample step    *** */
            if (!((t + 1) % WIt))
            {
               printf (" Starting L-sampler iteration %d\n", t + 1);
               fflush (stdout);
            }
            if (Scan)                   /* permute loci for L-sampler scan */
            {
               permute (perm_loc, nloc + 1);
               for (j = 0; j <= nloc; ++j)
               {
                  jj = perm_loc[j];

                  do_lsmpl (nloc, jj, prob, y, ngenes);
               }
               for (h = 1; h <= 2; ++h)
                  for (j = 0; j <= nloc; ++j)
                     lsflag[h][j] = 1;
               for (c = 0; c < n_compnts; ++c)
               {
                  l_ratio[c][1] = 0.0;
                  l_ratio[c][2] = l_ratio[c][1];

                  for (h = 1; h <= 2; ++h)
                  {
                     for (j = 0; j < nloc; ++j)
                        lsflag[h][j] = 1;
                     l_ratio[c][h] +=
                        diff_seg_prob (nloc, c, h, lsflag[h],
                                       temp_log_prob_seg);
                     cumulik[c][h] += exp (l_ratio[c][h]);
                  }
                  ++scounts[c];
               }
            }
            else
            {                           /* random locus, by name */
               jj = intran (0, nloc);
               /* adjust adjacent intervals out of l-ratio */
               /* note: relevant intervals may depend on h */
               for (h = 1; h <= 2; ++h)
                  for (j = 0; j <= nloc; ++j)
                     lsflag[h][j] = 0;

               for (h = 1; h <= 2; ++h)
               {
                  j = revchro[0][jj];   /* position under sim (0) */
                  if (((j == 0) && (lpflag[1][h][j] != 0)
                       || (lpflag[2][h][j] != 0))
                      || ((j == nloc) && (lpflag[1][h][j - 1] != 0)
                          || (lpflag[2][h][j - 1] != 0))
                      || ((lpflag[1][h][j] != 0) || (lpflag[2][h][j] != 0)
                          || (lpflag[1][h][j - 1] != 0)
                          || (lpflag[2][h][j - 1] != 0)))
                  {
                     if (j > 0)
                        lsflag[h][j - 1] = 1;
                     if (j < nloc)
                        lsflag[h][j] = 1;
                     if (chromo[0][j] != chromo[h][j])
                     {
                        j = revchro[h][jj];
                        if (j > 0)
                           lsflag[h][j - 1] = 1;
                        if (j < nloc)
                           lsflag[h][j] = 1;
                     }
                  }
                  for (c = 0; c < n_compnts; ++c)
                     l_ratio[c][h] -= diff_seg_prob (nloc, c, h, lsflag[h],
                                                     temp_log_prob_seg);
               }
               do_lsmpl (nloc, jj, prob, y, ngenes);
               for (h = 1; h <= 2; ++h)
                  for (c = 0; c < n_compnts; ++c)
                     l_ratio[c][h] += diff_seg_prob (nloc, c, h, lsflag[h],
                                                     temp_log_prob_seg);
               for (c = 0; c < n_compnts; ++c)
               {
                  for (h = 1; h <= 2; ++h)
                     cumulik[c][h] += exp (l_ratio[c][h]);
                  ++scounts[c];
               }
            }
         }
         else
         {                              /* ***    Do M-sample step    *** */
            if (!((t + 1) % WIt))
            {
               printf (" Starting M-sampler iteration %d\n", t + 1);
               fflush (stdout);
            }
            if (Scan)                   /* permute meioses for M-sampler scan */
            {
               permute (perm_sw, nsw);
               for (j = 0; j < nsw; ++j)
               {
                  l = perm_sw[j] + 1;

                  i = (meiosis[l].ind).index;
                  sx = meiosis[l].sex;
                  c = nghd[i].component - 1;

                  do_msmpl (nloc, c, i, sx, prob, y, ngenes);
               }
               for (h = 1; h <= 2; ++h)
                  for (j = 0; j <= nloc; ++j)
                     lsflag[h][j] = 1;
               for (c = 0; c < n_compnts; ++c)
               {
                  l_ratio[c][1] = 0.0;
                  l_ratio[c][2] = l_ratio[c][1];

                  for (h = 1; h <= 2; ++h)
                  {
                     for (j = 0; j < nloc; ++j)
                        lsflag[h][j] = 1;
                     l_ratio[c][h] += diff_seg_prob (nloc, c, h, lsflag[h],
                                                     temp_log_prob_seg);
                     cumulik[c][h] += exp (l_ratio[c][h]);
                  }
                  ++scounts[c];
               }
            }
            else
            {                           /*choose a random meiosis */
               l = intran (0, nsw - 1) + 1;
               i = (meiosis[l].ind).index;
               sx = meiosis[l].sex;
               c = nghd[i].component - 1;

               /* adjust meiosis to be updated out of l_ratio */

               for (h = 1; h <= 2; ++h)
                  l_ratio[c][h] -= lik_rat (nloc, i, sx, h);

               /*  do-msmpl */
               do_msmpl (nloc, c, i, sx, prob, y, ngenes);

               /* put updated meiosis back in to l_ratio and
                * update the cumulative ratio                */

               for (h = 1; h <= 2; ++h)
               {
                  l_ratio[c][h] += lik_rat (nloc, i, sx, h);
                  cumulik[c][h] += exp (l_ratio[c][h]);
               }
               ++scounts[c];
            }
         }
         if (!((t + 1) % ScIt))
         {                              /* score tot log-probs & recombs */
            for (c = 0; c < n_compnts; ++c)
            {
               for (j = 0; j <= nloc; ++j)
                  tot_log_prob_pen[c][j] += log (prob[c][j]);

               full_seg_prob (nloc, c, recounts, temp_log_prob_seg);
               for (j = 0; j < nloc; ++j)
                  tot_log_prob_seg[c][j] += temp_log_prob_seg[c][j];

               for (h = 1; h <= 2; ++h)
                  for (j = 0; j < nloc; ++j)
                     lsflag[h][j] = 1;

               (void) diff_seg_prob (nloc, c, 1, lsflag[1],
                                     temp2_log_prob_seg);
               for (j = 0; j < nloc; ++j)
                  tot_log_left_seg[c][j] += temp2_log_prob_seg[c][j] +
                     temp_log_prob_seg[c][j];

               (void) diff_seg_prob (nloc, c, 2, lsflag[2],
                                     temp2_log_prob_seg);
               for (j = 0; j < nloc; ++j)
                  tot_log_rght_seg[c][j] += temp2_log_prob_seg[c][j] +
                     temp_log_prob_seg[c][j];
            }
            ++tot_count;
         }
      }

      /* now we are done with this simulation/estimation for this trait locus
         position  */
      /* output results for this simulation/estimation set */

      for (c = 0; c < n_compnts; ++c)
      {
         printf ("\n Component %d:\n\n", c + 1);
         if (scounts[c] == 0)
            printf ("    No likelihood ratios because sampling count = 0\n\n");
         else
         {
            leftratio = cumulik[c][1] / scounts[c];
            rightratio = cumulik[c][2] / scounts[c];
            printf ("    Cumulative likelihood ratios (sampling count %d)\n",
                    scounts[c]);
            printf ("       estimated ratios:   left:sim = %6.4f,"
                    "  right:sim = %6.4f\n\n", leftratio, rightratio);
         }
         /* Fill in the matrices of likehood ratios */
         /* initialization first */

         for (j = 0; j < NTPos; ++j)
            likeratios[c][k][j] = 0.0;
         if (k == 0)
         {
            likeratios[c][0][NTPos - 1] = leftratio;
            likeratios[c][0][1] = rightratio;
         }
         else if (k == NTPos - 1)
         {
            likeratios[c][k][k - 1] = leftratio;
            likeratios[c][k][0] = rightratio;
         }
         else
         {
            likeratios[c][k][k - 1] = leftratio;
            likeratios[c][k][k + 1] = rightratio;
         }

         printf ("    Average total log-probability of segregations:\n");
         printf ("       first locus (constant) %8.3f, and by map interval:"
                 "\n", -nns[c + 1] * log (2.0));
         printf ("       interval        left          sim       right\n");
         for (j = 0; j < nloc; ++j)
            printf ("            %3d    %8.3f     %8.3f    %8.3f\n",
                    j + 1, tot_log_left_seg[c][j] / tot_count,
                    tot_log_prob_seg[c][j] / tot_count,
                    tot_log_rght_seg[c][j] / tot_count);
         printf ("\n");
         printf ("    Average total log-probability of penetrances,"
                 " by locus:");
         for (j = 0; j <= nloc; ++j)
         {
            if (!(j % 8))
               printf ("\n      ");
            if (chromo[0][j] == 0)
               printf ("  trait  ");
            else
               printf (" mark%2d  ", chromo[0][j]);
         }
         for (j = 0; j <= nloc; ++j)
         {
            if (!(j % 8))
               printf ("\n    ");
            printf (" %8.3f", tot_log_prob_pen[c][chromo[0][j]] / tot_count);
         }
         printf ("\n\n");
         fflush (stdout);

         /* print out recombination counts  */
         printf ("    Recombination counts for map intervals:\n");
         printf ("       intv   M: total    recomb   non-rec "
                 "  F: total    recomb   non-rec\n");
         for (j = 0; j < nloc; ++j)
         {
            printf ("       %4d ", j + 1);
            for (i = 0; i < 6; ++i)
            {
               if (i == 3)
                  printf (" ");
               printf ("%10d", recounts[c][j][i]);
            }
            printf ("\n");
         }
         printf ("\n");
         fflush (stdout);
      }
   }                                    /* ****    end of BIG loop   **** */

   /* now we are done with simulating for all nloc+2 simulation sets */

   printf ("\n\n ESTIMATED LODS SCORES\n");

   /* Call like_eigen */
   printf ("\n    The largest eigenvalue is:\n");
   for (c = 0; c < n_compnts; ++c)
   {
      like_eigenvalue = lik_eigen (likeratios[c], likes, NTPos);
      printf ("       component %d %8.4f\n", c + 1, like_eigenvalue);
      for (i = 0; i < NTPos - 1; ++i)
         lods[c][i] = log10 (likes[0] / likes[i + 1]);
   }
   printf ("\n    LOD scores:\n");
   printf ("       trait position");
   for (c = 0; c < n_compnts; ++c)
      printf ("  cmpnt%3d", c + 1);
   printf ("\n");
   for (i = 0; i < NTPos - 1; ++i)
   {
      printf ("        %7d      ", i + 1);
      for (c = 0; c < n_compnts; ++c)
         printf ("  %8.4f", lods[c][i]);
      printf ("\n");
   }
   printf ("\n");

   /* Free the memory */

   for (i = 0; i < n_compnts; ++i)
   {
      for (j = 0; j < NTPos; ++j)
      {
         free ((void *) likeratios[i][j]);
         likeratios[i][j] = NULL;
      }
      free ((void **) likeratios[i]);
      likeratios[i] = NULL;
      free ((void *) lods[i]);
      lods[i] = NULL;
   }
   free ((void ***) likeratios);
   likeratios = NULL;
   free ((void **) lods);
   lods = NULL;
   free ((void *) likes);
   likes = NULL;

   /* Free up allocated memory */

   free_y (&y, nloc);
   locus_peel_free ();
   free_allocate_loci ();
   free_ld_prob (&scounts, &recounts, &lsflag, &l_ratio, &prob, &cumulik,
                 &temp_log_prob_seg, &temp2_log_prob_seg, &tot_log_prob_seg,
                 &tot_log_prob_pen, &tot_log_left_seg, &tot_log_rght_seg);
   free_lpflag (2);
   free_logprob ();
   free_msmpl ();
   free_gen_pen ();
   free_seg_pen (nloc + 1);

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }
   if (nns)
   {
      free (nns);
      nns = NULL;
   }
   if (qdex)
   {
      free (qdex);
      qdex = NULL;
   }
   if (trtphen)
   {
      free (trtphen);
      trtphen = NULL;
   }
   exit (EXIT_SUCCESS);
}
