
/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/
/*     parsedrop.h:  header file for Pars/drop_pars()                   */
/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

#ifndef PARSEDROP_H             
#define PARSEDROP_H


struct tlist_node {                     /*Tlist_node:  linked or        */
    int id;                             /*unlinked/varying position     */
    int lnkd;                           /*trait to be simulated         */
    struct tlist_node * p_next;
                  };
typedef struct tlist_node TLIST;
typedef TLIST * TLIST_PTR;
static TLIST_PTR first_tlist_node = NULL;


struct prntmap_node {                  /*Prntmap_node:  marker and trait*/
    int chrom_no;                      /*loci for this chromosome       */
    int no_mark;
    int no_traits;
    struct prntmap_locus * p_locus;
    struct prntmap_node * p_next;
                    };
typedef struct prntmap_node PRNTMAP;
typedef PRNTMAP * PRNTMAP_PTR;
static PRNTMAP_PTR first_prntmap_node[3] = {NULL, NULL, NULL};

struct prntmap_locus {                 /*Prntmap_locus:  next marker or */
    int id;                            /*trait for this chromosome and  */
    int mark_id;                       /*distance                       */
    int locus_typ; 
    double dist;                      
    double loc_dist;                
    struct prntmap_locus * p_prev;
    struct prntmap_locus * p_next;
                     };
typedef struct prntmap_locus PRNTLOC;
typedef PRNTLOC * PRNTLOC_PTR;

                                       /*function prototypes            */

static SIM_PTR     nxt_sim_chrom(int *);
static DIST_PTR    nxt_mapd_trait(int *,int,SIM_PTR),
                   nxt_mapd_mark(int *,int,SIM_PTR);
static UNLNKD_PTR  nxt_unlnkd_trait(int *);
static TLIST_PTR   nxt_trait(int *);
static int         nxt_mark(int *,SIM_PTR);
static PRNTLOC_PTR nxt_prntmap_locus(int *,PRNTMAP_PTR);
static PRNTMAP_PTR nxt_prntmap(int *,int);
static int         test_sim(void),              
                   test_mark_map(SIM_PTR, int *),      
                   test_trait_map(SIM_PTR, int *),
                   test_trait_list(int,int),
                   test_mark_loci(SIM_PTR,int), test_trait_loci(int),
                   test_trait_vars(int),        test_trait_genotyp(int),
                   test_trait_dist(PRNTMAP_PTR,PRNTLOC_PTR),
                   test_seeds(void),            test_oseeds(void);
static PRNTMAP_PTR add_prntmap(SIM_PTR,int);
static void        add_only_mark_locus(PRNTMAP_PTR),
                   add_mark_locus(PRNTMAP_PTR,DIST_PTR),
                   add_trait_locus(PRNTMAP_PTR,DIST_PTR),
                   fix_locus_dist(PRNTMAP_PTR,PRNTLOC_PTR),
                   prnt_map(PRNTMAP_PTR, int),  
                   prnt_mark_freqs(SIM_PTR),
                   prnt_trait_freqs(PRNTMAP_PTR),
                   prnt_trait_pars(PRNTMAP_PTR),
                   prnt_unlnkd_freqs(void),
                   prnt_unlnkd_pars(void),
                   fmt_trait_freqs(char *,int,int),
                   fmt_trait_pars(char *,int),
                   alloc_recomb(void),          alloc_trait(void),
                   mak_recomb(PRNTLOC_PTR, int, int),
                   mak_no_recomb(UNLNKD_PTR, int, int),
                   mak_trait(int, int, int),
                   mak_freq(PRNTMAP_PTR, PRNTLOC_PTR, int, int);
static void        for_chrom(int),              on_chrom(int),
                   for_chrom_mark(int,int);

#endif  /*PARSEDROP_H*/
