
#include "msample.h"

/* routine for the interference Metropolis-Hastings */

/* find_patt determines recombination pattern from the segs */
/*  segs may be in chromosomal (1) or label (0) order       */

void  find_patt (rpt, segs, nl, id, rvc)
     int  *rpt, *segs, nl, id, *rvc;
            /* rpt will be the recomb pattern, segs is the meiosis vector */
            /* id=1, segs is in chromo order, 
               0 is in label order, rvc gives position, 
               nl=number of marker loci (trait also included)  */
{
   int   j, ss[NLOC];
   if (id == 1)
      for (j = 0; j <= nl; ++j)
         ss[j] = segs[j];
   else
      for (j = 0; j <= nl; ++j)
         ss[j] = segs[rvc[j]];
   for (j = 1; j <= nl; ++j)
   {
      rpt[j - 1] = 1;
      if (ss[j - 1] == ss[j])
         rpt[j - 1] = 0;
   }
   return;
}

/* binscore changes this to integer index  */

int   binscore (rpt, nl)
     int  *rpt, nl;
{
   int   j, ssj;
   ssj = 0;
   for (j = 0; j < nl; ++j)
      ssj = 2 * ssj + rpt[j];

   return (ssj);
}

/* interference patterns for jv-pedigree example */

void  make_patterns_jv (nn, patt_rat)
     int   nn;
     double *patt_rat;
{
   int   j;

   for (j = 0; j <= nn; ++j)
      patt_rat[j] = 0.0;

   patt_rat[0] = 0.854;                     /* INTERFERENCE PATTERN2 */
   patt_rat[1] = patt_rat[8] = 0.977;
   patt_rat[2] = patt_rat[4] = 1.465;
   patt_rat[9] = patt_rat[5] = patt_rat[10] = 1.953;

/*    patt_rat[0]=0.597;                     INTERFERENCE PATTERN 1
   patt_rat[1] = patt_rat[8] = 1.194;
   patt_rat[2] = patt_rat[4] = 1.221;
   patt_rat[6] = patt_rat[9] = 0.212;
   patt_rat[5] = patt_rat[10] = 4.672;
   patt_rat[3] = patt_rat[12] = 0.106;
   patt_rat[13] = patt_rat[11] = 0.422;
 */

   for (j = 16; j <= nn; ++j)
      patt_rat[j] = patt_rat[j - 16];
   return;
}

void  read_patterns (fp, nn, nmax, patt_rat)
     FILE *fp;
     int   nn, nmax;
     double *patt_rat;
{
   int   j, pd, bd;
   double cc;
   for (j = 0; j < nmax; ++j)
      patt_rat[j] = 0.0;

   for (j = 0; j < nn; ++j)
   {
      fscanf (fp, "%d %d %lf %lf %lf", &pd, &bd, &cc, &cc, &cc);
      patt_rat[pd] = cc;
   }
   for (j = nmax / 2; j < nmax; ++j)
      patt_rat[j] = patt_rat[j - nmax / 2];
   return;
}
