#include "lsample.h"
#include "msample.h"

/* even newer version of new-pen, which does everything in
 * one pass through the observed individuals which is much
 * faster (SCH, June 2001)  */

static int *group, *first, *next, *alleles[2], *gpflag, *gpsize, **obslist;
static int **cpstart, *gplist, *gplist1, ngroups;
static double *pp[2];

double seg_pen2 (int c, int ***y, double *freq, int locus, int ngenes, int *err)
{
   int  *genes[2];
   genes[0] = gg0[locus];
   genes[1] = gg1[locus];

   if (loci[locus].locus_type == MARKER_LOCUS)
      return (seg_pen (c, genes, y, freq, locus, ngenes, err));
   else
      return (gen_pen (c, genes, y, freq, locus, ngenes, err));
}

/* Allocate storage for seg_pen.  Should be called once only before
 * any seg_pen calls are made */
void  alloc_seg_pen (int ngenes, int nloc, int ***y)
{
   int   i, c, n, loc;
   Ind  *cpind;
   Indlist *iptr;

   if (CHK_PEEL (TRACE_LEVEL_1))
      (void) printf ("In alloc_seg_pen(%d)\n", ngenes);
   first = malloc (ngenes * 9 * sizeof (int));
   alloc_chk (WHERE, first);
   next = first + ngenes;                   /* First and next keep track of 
                                             * group members */
   group = next + ngenes;                   /* Group membership for each gene */
   gpflag = group + ngenes;                 /* Whether a group has 2 or 1 
                                             * possible states */
   gpsize = gpflag + ngenes;                /* Size of group */
   alleles[0] = gpsize + ngenes;            /* Allele allocations for each gene */
   alleles[1] = alleles[0] + ngenes;
   gplist = alleles[1] + ngenes;            /* List of active groups */
   gplist1 = gplist + ngenes;               /* Inverse of gplist */
   pp[0] = malloc (2 * ngenes * sizeof (double));  /* Probability of group */
   alloc_chk (WHERE, pp[0]);
   pp[1] = pp[0] + ngenes;
   /* Build up a list of observed individuals for each locus and each component.
    * Avoids cycling through entire pedigree each time */
   obslist = malloc (nloc * 2 * sizeof (void *));  /* List of observed
                                                      individuals for each locus
                                                      sorted by component */
   alloc_chk (WHERE, obslist);
   cpstart = obslist + nloc;                /* Starting location in obslist for
                                             * a given locus and component */
   cpstart[0] = malloc (nloc * (n_compnts + 1) * sizeof (int));
   for (i = 1; i < nloc; i++)
      cpstart[i] = cpstart[i - 1] + n_compnts + 1;
   for (loc = 0; loc < nloc; loc++)
   {
      obslist[loc] = 0;
      if (loci[loc].locus_type == MARKER_LOCUS)
      {
         for (n = c = 0; c < n_compnts; c++)
         {
            cpstart[loc][c] = n;
            iptr = cmplist[c + 1].people;
            while (iptr->next != NULL)
            {
               iptr = iptr->next;
               cpind = iptr->current;
               i = cpind->index;
               if (y[0][loc][i])
                  n++;
            }
         }
         cpstart[loc][c] = n;
         if (n)
         {
            obslist[loc] = malloc (n * sizeof (int));
            alloc_chk (WHERE, obslist[loc]);
            for (n = c = 0; c < n_compnts; c++)
            {
               iptr = cmplist[c + 1].people;
               while (iptr->next != NULL)
               {
                  iptr = iptr->next;
                  cpind = iptr->current;
                  i = cpind->index;
                  if (y[0][loc][i])
                     obslist[loc][n++] = i;
               }
            }
         }
      }
   }
}

/* Free all memory allocated by alloc_seg_pen.  Should be called once;
 * subsequent calls will have no effect */
void  free_seg_pen (int nloc)
{
   int   i;

   if (first)
   {
      free (first);
      first = 0;
   }
   if (pp[0])
   {
      free (pp[0]);
      pp[0] = 0;
   }
   if (cpstart && cpstart[0])
   {
      free (cpstart[0]);
      cpstart[0] = 0;
   }
   if (obslist)
   {
      for (i = 0; i < nloc; i++)
         if (obslist[i])
            free (obslist[i]);
      free (obslist);
      obslist = 0;
   }
}

/* Change group g1 to group g2, link groups together, and switch allele 
 * allocations */
static void change_and_switch_group (int g1, int g2)
{
   int   k, k1;

   /* Change group and allele allocations for members of g1 */
   k1 = gpflag[g2 - 1] - 1;
   k = first[g1 - 1];
   while (next[k] >= 0)
   {
      group[k] = g2;
      alleles[k1][k] = alleles[k1 ^ 1][k];
      k = next[k];
   }
   group[k] = g2;
   alleles[k1][k] = alleles[k1 ^ 1][k];
   /* Add members of g1 to g2 list */
   next[k] = first[g2 - 1];
   first[g2 - 1] = first[g1 - 1];
   /* Update probabilities for group g2 */
   pp[k1][g2 - 1] *= pp[k1 ^ 1][g1 - 1];
   /* Update size of g2 */
   gpsize[g2 - 1] += gpsize[g1 - 1];
   gpsize[g1 - 1] = 0;
   /* Remove g1 from active list */
   k = gplist1[g1 - 1];
   k1 = gplist[k] = gplist[--ngroups];
   gplist1[k1] = k;
}

/* Change group g1 to group g2, link groups together, and swap allele 
   allocations */
static void change_and_swap_group (int g1, int g2)
{
   int   k, k1;

   /* Change group and allele allocations for members of g1 */
   k = first[g1 - 1];
   while (next[k] >= 0)
   {
      group[k] = g2;
      k1 = alleles[0][k];
      alleles[0][k] = alleles[1][k];
      alleles[1][k] = k1;
      k = next[k];
   }
   group[k] = g2;
   k1 = alleles[0][k];
   alleles[0][k] = alleles[1][k];
   alleles[1][k] = k1;
   /* Add members of g1 to g2 list */
   next[k] = first[g2 - 1];
   first[g2 - 1] = first[g1 - 1];
   /* Update log probabilities for group g2 */
   pp[0][g2 - 1] *= pp[0][g1 - 1];
   pp[1][g2 - 1] *= pp[1][g1 - 1];
   /* Update size of g2 */
   gpsize[g2 - 1] += gpsize[g1 - 1];
   gpsize[g1 - 1] = 0;
   /* Remove g1 from active list */
   k = gplist1[g1 - 1];
   k1 = gplist[k] = gplist[--ngroups];
   gplist1[k1] = k;
}

/* Change group g1 to group g2 and link groups together */
static void change_group (int g1, int g2)
{
   int   k, k1;

   /* Change group for members of g1 */
   k = first[g1 - 1];
   while (next[k] >= 0)
   {
      group[k] = g2;
      k = next[k];
   }
   group[k] = g2;
   /* Add members of g1 to g2 list */
   next[k] = first[g2 - 1];
   first[g2 - 1] = first[g1 - 1];
   k = gpflag[g2 - 1];
   /* Update probabilities for group g2 */
   if (k--)
      pp[k][g2 - 1] *= pp[k][g1 - 1];
   else
   {
      pp[0][g2 - 1] *= pp[0][g1 - 1];
      pp[1][g2 - 1] *= pp[1][g1 - 1];
   }
   /* Update size of g2 */
   gpsize[g2 - 1] += gpsize[g1 - 1];
   gpsize[g1 - 1] = 0;
   /* Remove g1 from active list */
   k = gplist1[g1 - 1];
   k1 = gplist[k] = gplist[--ngroups];
   gplist1[k1] = k;
}

double seg_pen (int c, int *genes[2], int ***y, double *freq, int locus,
                int ngenes, int *err)
{
   int   i, j, k, k1, s, s1, t, a1, a2, g1, g2, o1, o2, o1a, o2a;
   int  *obs, *y1, *y2;
   int   k_sw[] =
   {-1, 1, 1, -1, 2, -1, 3, -1, 2, 0, -1, -1, -1, -1, -1, -1};
   int   k1_sw[] =
   {-1, 1, 2, -1, 1, -1, 3, -1, 2, 0, -1, -1, -1, -1, -1, -1};
   double p, z;

   if (CHK_PEEL (TRACE_LEVEL_1))
      (void) printf ("\nIn seg_pen(%d,%p,%p,%p,%d,%d,%p)\n", c, genes, y,
                     freq, locus, ngenes, err);
   s1 = cpstart[locus][c];
   s = cpstart[locus][c + 1] - s1;
   /* If no observed individuals in component then return immediately */
   if (!s)
   {
      *err = 0;
      return 1.0;
   }
   *err = -1;
   /* Blank group array */
   (void) memset (group, 0, sizeof (int) * ngenes);
   /* Clear group counts */
   j = ngroups = 0;
   /* Dereference pointers for speed */
   y1 = y[0][locus];
   y2 = y[1][locus];
   obs = obslist[locus] + s1;
   while (s--)
   {
      i = *(obs++);                         /* Get next observed individual in 
                                             * component */
      o1 = y1[i];                           /* Get observed data */
      o2 = y2[i];
      a1 = genes[0][i] - 1;                 /* and the founder genes */
      a2 = genes[1][i] - 1;
      /* Is individual autozygous ? */
      if (a1 == a2)
      {
         /* Yes, Check homozygosity */
         if (o1 != o2)
            return 0.0;
         g1 = group[a1];
         /* Already in group ? */
         if (g1)
         {
            /* Yes, is group is fixed ? */
            k = gpflag[g1 - 1];
            if (k--)
            {
               /* Yes, Check allele matches ? */
               if (alleles[k][a1] != o1)
                  return 0.0;
            }
            else
            {
               /* No, find which alternate matches current allele */
               for (k = 0; k < 2; k++)
               {
                  o1a = alleles[k][a1];
                  if (o1a == o1)
                     break;
               }
               if (k == 2)
                  return 0.0;               /* No match * 
                                             * Fix group appropriately */
               gpflag[g1 - 1] = k + 1;
            }
         }
         else
         {
            /* No, create new group */
            first[j] = a1;
            gpflag[j] = gpsize[j] = 1;
            next[a1] = -1;
            pp[0][j] = freq[o1];            /* Add in contrib. to pp from this gene */
            alleles[0][a1] = o1;
            /* Add group to active list */
            gplist[ngroups] = j;
            gplist1[j] = ngroups++;
            /* Set group membership for a1 and increment group counter */
            group[a1] = ++j;
         }
      }
      else
      {
         /* Not autozygous, check if in existing group */
         g1 = group[a1];
         g2 = group[a2];
         if (g1)
         {
            k = gpflag[g1 - 1];
            if (g2)
            {
               /* Both genes in groups, same group ? */
               if (g1 == g2)
               {
                  /* Yes, group fixed ? */
                  if (k--)
                  {
                     o1a = alleles[k][a1];
                     o2a = alleles[k][a2];
                     /* Yes, check match */
                     if ((o1a != o1 || o2a != o2) &&
                         (o1a != o2 || o2a != o1))
                        return 0.0;
                  }
                  else
                  {
                     /* No, find which alternate matches */
                     for (k1 = k = 0; k < 2; k++)
                     {
                        o1a = alleles[k][a1];
                        o2a = alleles[k][a2];
                        if ((o1a != o1 || o2a != o2) &&
                            (o1a != o2 || o2a != o1))
                           k1 |= 2 - k;
                     }
                     if (k1 == 3)
                        return 0.0;         /* No match */
                     gpflag[g1 - 1] = k1;   /* Fix group */
                  }
               }
               else
               {
                  /* Genes are in different groups. Groups fixed ? */
                  k1 = gpflag[g2 - 1];
                  if (k)
                  {
                     o1a = alleles[k - 1][a1];
                     if (k1)
                     {
                        o2a = alleles[k1 - 1][a2];
                        /* Both groups fixed, check allocations match */
                        if ((o1a != o1 || o2a != o2) &&
                            (o1a != o2 || o2a != o1))
                           return 0.0;
                     }
                     else
                     {
                        /* Only first group fixed.  Check which possible
                         * allocations for second group match */
                        for (k1 = 0; k1 < 2; k1++)
                        {
                           o2a = alleles[k1][a2];
                           if ((o1a == o1 && o2a == o2) ||
                               (o1a == o2 && o2a == o1))
                              break;
                        }
                        if (k1++ == 2)
                           return 0.0;      /* No match */
                     }
                  }
                  else if (k1)
                  {
                     /* Only second group fixed.  Check which possible
                      * allocations for first group match */
                     o2a = alleles[k1 - 1][a2];
                     for (k = 0; k < 2; k++)
                     {
                        o1a = alleles[k][a1];
                        if ((o1a == o1 && o2a == o2) ||
                            (o1a == o2 && o2a == o1))
                           break;
                     }
                     if (k++ == 2)
                        return 0.0;         /* No match */
                  }
                  else
                  {
                     /* No groups fixed.  See which allocations match up */
                     for (k1 = k = 0; k < 4; k++)
                     {
                        o1a = alleles[(k & 2) >> 1][a1];
                        o2a = alleles[k & 1][a2];
                        if ((o1a == o1 && o2a == o2) ||
                            (o1a == o2 && o2a == o1))
                           k1 |= 1 << k;
                     }
                     if (!k1)
                        return 0.0;         /* No match */
                     /* Convert from k to appropriate gpflag entries */
                     k = k_sw[k1];
                     k1 = k1_sw[k1];
                     if (k < 0 || k1 < 0)
                     {
                        fprintf (stderr, "Internal Error - bad code\n");
                        exit (-1);
                     }
                  }
                  /* Are allocation flags the same ? */
                  if (k != k1)
                  {
                     /* No, so we must flip the allocations for all members of
                      * one of the groups (the smallest) */
                     if (gpsize[g1 - 1] < gpsize[g2 - 1])
                     {
                        gpflag[g2 - 1] = k1;
                        change_and_switch_group (g1, g2);
                     }
                     else
                     {
                        gpflag[g1 - 1] = k;
                        change_and_switch_group (g2, g1);
                     }
                  }
                  else
                  {
                     /* Yes, change group membership for smallest group */
                     if (gpsize[g1 - 1] < gpsize[g2 - 1])
                     {
                        if (k < 3)
                        {
                           gpflag[g2 - 1] = k;
                           change_group (g1, g2);
                        }
                        else
                        {
                           gpflag[g2 - 1] = 0;
                           change_and_swap_group (g1, g2);
                        }
                     }
                     else
                     {
                        if (k < 3)
                        {
                           gpflag[g1 - 1] = k;
                           change_group (g2, g1);
                        }
                        else
                        {
                           gpflag[g1 - 1] = 0;
                           change_and_swap_group (g2, g1);
                        }
                     }
                  }
                  g1 = group[a1];
                  k = gpflag[g1 - 1] - 1;
               }
            }
            else
            {
               /* Only a1 in group, a2 is new */
               if (k)
               {
                  /* Group for a1 is fixed, check matches */
                  o1a = alleles[k - 1][a1];
                  if (o1a == o1)
                     alleles[k - 1][a2] = o2;
                  else if (o1a == o2)
                     alleles[k - 1][a2] = o1;
                  else
                     return 0.0;            /* No match */
               }
               else
               {
                  /* Group for a1 not fixed, check both allocations for matches */
                  for (k1 = k = 0; k1 < 2; k1++)
                  {
                     o1a = alleles[k1][a1];
                     if (o1a == o1)
                        alleles[k1][a2] = o2;
                     else if (o1a == o2)
                        alleles[k1][a2] = o1;
                     else
                        k |= 2 - k1;
                  }
                  if (k == 3)
                     return 0.0;            /* No match */
                  gpflag[g1 - 1] = k;
               }
               /* Add a2 to group */
               t = g1 - 1;
               next[a2] = first[t];
               first[t] = a2;
               group[a2] = g1;
               gpsize[t]++;
               /* Add in contrib. to pp from gene a2 */
               if (k--)
                  pp[k][t] *= freq[alleles[k][a2]];
               else
                  for (k = 0; k < 2; k++)
                     pp[k][t] *= freq[alleles[k][a2]];
            }
         }
         else if (g2)
         {
            /* Only a2 in group, a1 is new */
            k1 = gpflag[g2 - 1];
            if (k1)
            {
               /* Group for a2 is fixed, check matches */
               o1a = alleles[k1 - 1][a2];
               if (o1a == o1)
                  alleles[k1 - 1][a1] = o2;
               else if (o1a == o2)
                  alleles[k1 - 1][a1] = o1;
               else
                  return 0.0;               /* No match */
            }
            else
            {
               /* Group for a2 not fixed, check both allocations for matches */
               for (k1 = k = 0; k < 2; k++)
               {
                  o2a = alleles[k][a2];
                  if (o2a == o1)
                     alleles[k][a1] = o2;
                  else if (o2a == o2)
                     alleles[k][a1] = o1;
                  else
                     k1 |= 2 - k;
               }
               if (k1 == 3)
                  return 0.0;               /* No match */
               gpflag[g2 - 1] = k1;
            }
            /* Add a1 to group */
            t = g2 - 1;
            next[a1] = first[t];
            first[t] = a1;
            group[a1] = g2;
            gpsize[t]++;
            /* Add in contrib. to pp from gene a1 */
            if (k1--)
               pp[k1][t] *= freq[alleles[k1][a1]];
            else
               for (k1 = 0; k1 < 2; k1++)
                  pp[k1][t] *= freq[alleles[k1][a1]];
         }
         else
         {                                  /* A first for both genes */
            if (o1 != o2)
            {                               /* Heterozygote */
               alleles[0][a1] = alleles[1][a2] = o1;
               alleles[1][a1] = alleles[0][a2] = o2;
               gpflag[j] = 0;
               z = freq[o1] * freq[o2];
               pp[0][j] = pp[1][j] = z;
            }
            else
            {                               /* Homozygote */
               alleles[0][a1] = alleles[0][a2] = o1;
               gpflag[j] = 1;
               z = freq[o1];
               pp[0][j] = z * z;
            }
            /* Create new group */
            first[j] = a1;
            next[a1] = a2;
            next[a2] = -1;
            gpsize[j] = 2;
            /* Add to active list */
            gplist[ngroups] = j;
            gplist1[j] = ngroups++;
            group[a1] = group[a2] = ++j;
         }
      }
   }
   /* All allocations have been made.  Go through active group list
    * and accumulate the probability */
   p = 1.0;
   *err = 0;
   for (j = 0; j < ngroups; j++)
   {
      k = gplist[j];
      k1 = gpflag[k];
      if (k1--)
      {                                     /* Group is fixed 
                                             * (only 1 possible allocation) */
         p *= pp[k1][k];
      }
      else
      {                                     /* 2 possible allocations for group
                                             */
         p *= (pp[0][k] + pp[1][k]);
      }
   }
   if (CHK_PEEL (TRACE_LEVEL_1))
      (void) printf ("Returning like = %g, p = %g\n", log (p), p);
   return p;
}
