#include "msample.h"

/* set_chromosomes(), print_chromosome(), set_lpflag(), 
 * alloc_lpflag(), free_lpflag()        
 */
void  set_chromosomes (int nl, double *map, int nh, int *l0, double *r0,
                       int ix)
/* routine to set the array of chromosome recombinations and orders for
   simulations and likelihood evaluations;
   nl = number of marker loci, map = recombination fractions between;
   nh = number of alternative positions for disease -- 0-one is simulation one
   ix = sex of map -- 0=avg, 1=male, 2=female
   (r0,l0) vector of pairs giving position of disease; 
   l0 is marker locus to the left and recombination r0 is to that marker; 
   note r0 <= r(l0).
   (if disease extreme left, l0=0, and r0 is recom to locus 1), 

   Also to set flags for log-prob diffs, so more efficient.  EAT 12/00
 */
{
   int   i, l;

   for (i = 0; i <= nh; ++i)                /* set chromosome i */
   {
      if (l0[i] == 0)
      {                                     /* for trait first on chromosome */
         chromo[i][0] = 0;
         xecomo[ix][i][0] = r0[i];
         revchro[i][0] = 0;
         for (l = 1; l <= nl; ++l)
         {
            chromo[i][l] = l;
            xecomo[ix][i][l] = map[l];
            revchro[i][l] = l;
         }
      }
      else if (l0[i] == nl)
      {                                     /* for trait last on chromosome */
         for (l = 1; l <= nl; ++l)
         {
            chromo[i][l - 1] = l;
            xecomo[ix][i][l - 1] = map[l];
            revchro[i][l] = l - 1;
         }
         chromo[i][nl] = 0;
         xecomo[ix][i][nl - 1] = r0[i];
         xecomo[ix][i][nl] = .5;
         revchro[i][0] = nl;
      }
      else
      {                                     /* for trait between marker loci 
                                               on chromosome */
         for (l = 0; l < l0[i] - 1; ++l)
         {
            chromo[i][l] = l + 1;
            xecomo[ix][i][l] = map[l + 1];
            revchro[i][l + 1] = l;
         }
         l = l0[i] - 1;
         chromo[i][l] = l + 1;
         xecomo[ix][i][l] = r0[i];
         revchro[i][l + 1] = l;
         ++l;
         chromo[i][l] = 0;
         revchro[i][0] = l;
         xecomo[ix][i][l] = (map[l] - r0[i]) / (1.0 - 2.0 * r0[i]);

         for (l = l0[i] + 1; l <= nl; ++l)
         {
            chromo[i][l] = l;
            xecomo[ix][i][l] = map[l];
            revchro[i][l] = l;
         }
      }
      /* code moved to set_lpflag() */
   }
}

void  print_chromosome (int nl, int hh, int ix)
{
   int   j;

   for (j = 0; j < nl; ++j)
      printf ("%3d %6.3f", chromo[hh][j], xecomo[ix][hh][j]);
   printf ("%3d\n", chromo[hh][nl]);
   fflush (stdout);

   return;
}
void  set_lpflag (int nl, int nh, int ix)
/*
 * Routine to set the sex-specific marker maps, and the trait locus 
 * position for jnt_auto, and lm_auto programs;       EAT Mar 00   
 *
 * set_maps() removed                                 MAJ Nov 00 
 *
 * Sets flags for log-prob diffs, so more efficient.  EAT DEC 00
 * 
 *
 * nl = number of marker loci
 * nh = number of alternative positions for disease -- 0-one is simulation one.
 * ix = sex of map 0=avg, 1=male, 2=female
 */
{
   int   i, jj, jr, jl, jsr, jsl;

   for (i = 1; i <= nh; ++i)                /* set chromosome i */
   {
      for (jj = 0; jj < nl; ++jj)
      {
         lpflag[ix][i][jj] = 0;
         jr = chromo[i][jj + 1];
         jsr = chromo[0][jj + 1];
         jl = chromo[i][jj];
         jsl = chromo[0][jj];
         if ((jr != jsr) || (jl != jsl) ||
             (xecomo[ix][i][jj] != xecomo[ix][0][jj]))
            lpflag[ix][i][jj] = 1;
      }
   }
   return;
}
void  alloc_lpflag (int nl, int nh)
{
/* mallocs lpflag [3][nh+1][nl]                         MAJ Sep 01 */

   int   i, j;

   lpflag = malloc (3 * sizeof (int **));
   alloc_chk (WHERE, lpflag);
   for (i = 0; i < 3; ++i)
   {
      lpflag[i] = malloc ((nh + 1) * sizeof (int *));
      alloc_chk (WHERE, lpflag[i]);
      for (j = 0; j <= nh; ++j)
      {
         lpflag[i][j] = malloc (nl * sizeof (int));
         alloc_chk (WHERE, lpflag[i][j]);
      }
   }
   return;
}
void  free_lpflag (int nh)
{
   int   i, j;

   if (!lpflag)
      return;
   for (i = 0; i < 3; ++i)
   {
      for (j = 0; j <= nh; ++j)
         free (lpflag[i][j]);
      free (lpflag[i]);
   }
   free (lpflag);
   lpflag = NULL;

   return;
}
