#include "nghds.h"

/***************************************************************** 
  Function to read pedigree data in from a file written in free format and 
  to form a data matrix for the pedigree. 
  This version uses the "genetic data columns" as fixed-effects indicators,
  and also reads in a quantitative data value "quant" -- becomes "x" 
  The nghds() and position() routines are then called
  
  Various checks on pedigree size etc. are carried out and the array "posit" 
  giving a one to one relationship between pedigree name and row index 
  number in the data matrix is also formed. The global variables 
  "pedsize" and "namestring" are also evaluated here.
  
  Modified out of input_data() in Quant library;  6/8/93. (EAT).
  
*****************************************************************/

void  input_ped (FILE * fp, int nn, int **dodo, double **quant)
{
   int   i, j, k, n, *names;

   for (i = 0; i < PedSize; ++i)
   {
      for (j = 0; j < NInts; ++j)
         if (k = SelInt[j])
            n = fscanf (fp, "%d", &data[i][k]);
         else
            n = fscanf (fp, "%*d");

      for (j = 0, k = 0; j < NReals; ++j)
         if (SelTrait[j] == TRUE)
         {
            ++k;
            n = fscanf (fp, "%lf", &quant[k - 1][i]);
         }
         else
            n = fscanf (fp, "%*lf");

      for (j = 0; j < nn; ++j)
         n = fscanf (fp, "%d", &dodo[j][i]);

      if (n == EOF)
         break;

      data[i][0] = i;
   }

   fprintf (stdout, " Pedigree data file records:  ");
   if (NReals > 1)
      fprintf (stdout, "%d integers and %d real numbers\n\n", NInts, NReals);
   else if (NReals == 1)
      fprintf (stdout, "%d integers and 1 real number\n\n", NInts);
   else
      fprintf (stdout, "%d integers\n\n", NInts);

   fprintf (stdout, "% d individuals read\n\n", i);
   fflush (stdout);
   if (i < PedSize)
   {
      give_warn (ERR_nex, PedSize);
      if (i < 3)
         give_up (ERR_siz, "input_ped");
   }
   else
   {
      n = fscanf (fp, "%*s");
      if (n != EOF)
         give_warn (ERR_urd, "input_ped");
   }
   pedsize = i;

   names = (int *) malloc (pedsize * sizeof (int));
   alloc_chk (WHERE, names);

   for (i = 0; i < pedsize; ++i)
      names[i] = data[i][1];                /* Extracting the column of names */
   namestring = intmax (names, 0, pedsize);

   free ((void *) names);

   if (namestring > MaxNames)
      give_up (ERN_mxn, namestring, MaxNames);

   alloc_pos ();

   /* Now set neighbourhoods */

   position (data, posit);
   nghds (data);
   cmplist = (Component *) malloc (MaxNCmps * sizeof (Component));
   alloc_chk (WHERE, cmplist);

   n_compnts = components ();

   if (n_compnts > 1)
      fprintf (stdout, " %d components found\n\n"
               " %d members are in the largest component\n\n",
               n_compnts, max_cpsize);
   return;
}
