
#include "pedchk.h"

/* opens and reads the pedigree file.  Stores the data in the pedchk    */
/* data structures.                                                     */
/* on return, leaves the keytable sorted in chronological order         */

int   chk_pedfil (void)
{
   FILE *pedF;

   if (!(pedF = open_ped ()))           /* opens pedigree file          */
      return !OK;

   allocPed ();                         /* allocates arrays             */

   if (!readPed (pedF))                 /* reads pedigree file          */
      return !OK;

   if (!checkPed (pedF))                /* reports pedigree size        */
      return !OK;
   fclose (pedF);

   allocKey ();                         /* allocates key table          */

   storKey ();                          /* stores names in key table    */

   sortNam ();                          /* sorts key table on name      */

   if (!checkNam ())                    /* checks for dupicate names    */
      return !OK;

   findParents ();                      /* stores ix's to parents in
                                         * key table */

   if (!checkParents ())                /* checks for missing parents   */
      return !OK;

   assignGender ();                     /* if requested, assigns gender */

   if (!checkGender ())                 /* checks parents for correct   */
      return !OK;                       /* gender                       */

   sortIx ();                           /* sorts key table on rec. no.  */

   assignChron ();                      /* finds chronological order
                                         * and stores no. in key table  */

   if (!checkChron ())                  /* checks for chronological     */
      return !OK;                       /* order found                  */

   sortChron ();                        /* sorts key table on chron no. */

   return OK;
}
