/*
 * Routine to generate a peeling sequence for a zero loop pedigree 
 * Original program;  Winter 1993 - Hannah Payne
 * Routine integrated into Nghds framework;  Jan 1994 (EAT)
 * Replaced fixed arrays by pointers etc.  (EAT) 
 *
 * Augment linked list of families by original paprs array , so that they 
 * can be put into the list in random order to produce different peeling 
 * sequences.   EAT;  April 1996. 
 *
 * Modified so final pivot is -1.  Will still only work for single
 * connected zero-loop pedigree, but can be done by components for 
 * multiple pedigrees.
 *              EAT;  March 1997. 
 */

#include "nghds.h"

typedef struct _Nucfam
{
   Ind  *mum, *dad, *pivot;
   int   done, numkids;
   struct _Indlist *kids;
}
Nucfam;

typedef struct _Famlist
{
   struct _Nucfam *current;
   struct _Famlist *next;
}
Famlist;

int   peel_seq (char out_file[], int r)  /* r=1 for random sequence */
{
   int  *degree;
   int **paprs, nfam;
   int  *permu;
   Famlist *families, *fptr;
   Nucfam *nucfam;
   Ind  *ipt;
   Indlist *offspring;
   FILE *fp;

   int   flag = 0, peel = 0, skip = 0;
   int   i, j;

   degree = (int *) malloc (pedsize * sizeof (int));
   for (j = 0; j < pedsize; j++)
      degree[j] = 0;

   paprs = (int **) malloc (pedsize * sizeof (int *));
   paprs[0] = (int *) malloc (pedsize * 2 * sizeof (int));
   paprs[1] = paprs[0] + pedsize;

   fp = fopen (out_file, "w");

   /*
    * Get parent pairs (parprs) - all pairs of parents that exist.
    * Paprs are labeled 0 to (nfam-1)     
    */

   nfam = 0;

   for (i = 0; i < pedsize; ++i)
   {
      ipt = &nghd[i];
      if ((ipt->pa->self) != UNKNOWN)
      {
         j = 0;
         while (j <= nfam)
         {
            if ((paprs[0][j] == ipt->pa->index)
                && (paprs[1][j] == ipt->ma->index))
               flag = 1;
            ++j;
         }
         if (flag != 1)
         {
            paprs[0][nfam] = ipt->pa->index;
            paprs[1][nfam] = ipt->ma->index;
            ++nfam;
         }
         flag = 0;
      }
   }
   /* If r=1; Set up a permutation of 1 to nfam to get random order */

   permu = (int *) malloc (nfam * sizeof (int));
   for (i = 0; i < nfam; ++i)
      permu[i] = i;
   if (r == 1)
      permute (permu, nfam);

   /* Put families into linked list  according to the permutation */

   families = (Famlist *) malloc (sizeof (Famlist));
   families->next = NULL;
   fptr = families;

/*
 * for (i = 0; i < nfam; ++i)
 * {
 *    fptr->next = (Famlist *) malloc(sizeof(Famlist));
 *    fptr->current = (Nucfam *) malloc(sizeof(Nucfam));
 *    nucfam = fptr->current;
 *    nucfam->dad = &nghd[paprs[0][permu[i]]];
 *    nucfam->mum = &nghd[paprs[1][permu[i]]];
 *    nucfam->kids = (Indlist *) malloc(sizeof(Indlist));
 *    fptr = fptr->next;
 *  }
 */

   i = 0;
   while (i < nfam)
   {
      fptr->current = (Nucfam *) malloc (sizeof (Nucfam));
      fptr->next = NULL;
      nucfam = fptr->current;
      nucfam->dad = &nghd[paprs[0][permu[i]]];
      nucfam->mum = &nghd[paprs[1][permu[i]]];
      nucfam->kids = (Indlist *) malloc (sizeof (Indlist));

      if (++i >= nfam)
         continue;

      fptr->next = (Famlist *) malloc (sizeof (Famlist));
      fptr = fptr->next;
   }
   fprintf (fp, "\n");

   /*
    * Create nuclear families.  For all parprs get corresponding kids and
    * keep track of the number of kids (numkids) in each nuclear family   
    *
    * Also; Find degree of each individual (by position index) 
    * Initialize done to be 0 for each nuclear family - 
    * the "done" index will become 1 if nuclear family is peeled
    */

   fptr = families;
   while (fptr->next != NULL)
   {
      nucfam = fptr->current;
      nucfam->numkids = 0;
      nucfam->done = 0;
      ++degree[nucfam->dad->index];
      ++degree[nucfam->mum->index];
      offspring = nucfam->kids;
      for (i = 0; i < pedsize; ++i)
      {
         ipt = &nghd[i];
         if ((nucfam->dad == ipt->pa) && (nucfam->mum == ipt->ma))
         {
            offspring->next = (Indlist *) malloc (sizeof (Indlist));
            offspring->current = ipt;
            ++degree[ipt->index];
            ++(nucfam->numkids);
            offspring = offspring->next;
         }
      }
      offspring->next = NULL;
      fptr = fptr->next;
   }

   /* Print out families */
   /* 
    * fptr = families;
    * while (fptr->next != NULL)
    * {
    *    nucfam = fptr->current;
    *    printf("\n %d %d  %d", nucfam->mum->self, nucfam->dad->self,
    *          nucfam->numkids);
    *    offspring = nucfam->kids;
    *    while (offspring->next != NULL)
    *   {
    *       printf("  %d ", offspring->current->self);
    *       offspring = offspring->next;
    *   }
    *    fptr = fptr->next;
    * }  
    * fflush(stdout);
    */

   /*
    * Loop to generate and print peeling sequence.  For each nuclear family,
    * check to see if its members' degrees all = 1 except for one individual
    * with degree > 1 (the pivot).  Then this family is peelable, and the
    * father, mother, pivot, #kids, kid1, kid2, ..., kidk are printed
    * (parpr, pivot, numkid, kid).  Then, the degree of all members in that
    * nuclear family are decreased by 1 and the loop looks for the next
    * peelable family. This process continues every family is peeled
    * (otherwise there must be a data error). 
    *
    * For the final family all individuals have degree 1, and the pivot is 0.
    * If this happens before the final family, pedigree is disconnected  
    */

   fptr = families;
   while (fptr->next != NULL)
   {
      nucfam = fptr->current;
      peel = 0;
      skip = 0;
      if (nucfam->done == 1)
         skip = 1;
      else
      {
         if (degree[nucfam->dad->index] > 1)
         {
            ++peel;
            nucfam->pivot = nucfam->dad;
         }
         if (degree[nucfam->mum->index] > 1)
         {
            ++peel;
            nucfam->pivot = nucfam->mum;
         }
         offspring = nucfam->kids;
         while (offspring->next != NULL)
         {
            ipt = offspring->current;
            if (degree[ipt->index] > 1)
            {
               ++peel;
               nucfam->pivot = ipt;
            }
            offspring = offspring->next;
         }
      }
      if ((peel <= 1) && (skip == 0))
      {
         fprintf (fp, " %d %d", nucfam->mum->self, nucfam->dad->self);
         --degree[nucfam->mum->index];
         --degree[nucfam->dad->index];
         if (peel == 1)
            fprintf (fp, "   %d  %d ", nucfam->pivot->self, nucfam->numkids);
         else
            fprintf (fp, "   -1  %d ", nucfam->numkids);
         offspring = nucfam->kids;
         while (offspring->next != NULL)
         {
            ipt = offspring->current;
            fprintf (fp, " %d ", ipt->self);
            --degree[ipt->index];
            offspring = offspring->next;
         }
         fprintf (fp, "\n");
         (nucfam->done) = 1;
         if (peel == 1)
            fptr = families;
         else
            break;
      }
      else
         fptr = fptr->next;
   }

   /* when falls out the bottom of this loop should be done  */

   fptr = families;
   while (fptr->next != NULL)
   {
      nucfam = fptr->current;
      if (nucfam->done != 1)
         give_up_where (WHERE, ERP_con);
      else
         fptr = fptr->next;
   }

   /*
    * Check to make sure degree = 0 for everyone after peeling sequence 
    * is done      
    */

   flag = 0;
   for (j = 0; j < pedsize; ++j)
      if (degree[j] != 0)
         ++flag;

   if (flag > 0)
      give_up_where (WHERE, ERP_nzd, flag);

   free ((void *) degree);
   fclose (fp);

   return 0;
}
