/************************************************************************
   programs to subtract fixed effects, both all, and all_bar_one,
   for both the univariate and bivariate cases.		
  
   Installed in Quant library 9/13/91   
  
   Univar and bivar versions combined; FfDex, Mufti, etc. fixed as arrays 
   for greater compatibility, and for multivariate case.
	5/27/92 (EAT)
  
**************************************************************************
  
   includes also a modifed version of SWG's out_gene(), allowing
   subtraction of major-gene effects.
   installed in Quant library  5/26/92
   modified for proportional genotype effects;  10/93; EAT
  
**************************************************************************
  
   includes also fixeffit() which estimates fixed effects
   -- used by bivar, and to be used now also by univar and mixd
   Installed in Quant library, 5/28/92
  
**************************************************************************/

#include "quant.h"

void  out_fixed (double *y, int l, int f)
/* subtract out all fixed effects trait (l+1) -- fixed-effect set f   */
/* NOTE: this is done with reference to ORIGINAL data vector x[l][]   */
{
   int   i, j;

   Mufti[f][0][1] = mu[f];
   for (i = 1; i <= np; ++i)
      y[i] = x[l][i];
   for (j = 0; j <= NFix; ++j)
      for (i = 1; i <= np; ++i)
         y[i] -= Mufti[f][j][fixdex[j][i]] * FfDex[l][j];
}

void  out_bar_one (double *a, double *y, double *xx, int j, int ll)
/* subtract out all effects, except those of class j; trait ll+1 */
/*
 * NOTE: modified, Aug 7, 1992, so that subtraction is relative 
 * to vector xx, which may NOT be original data -- could be out-gened 
 */
{
   int   i, l;

   for (i = 1; i <= np; ++i)
      y[i] = xx[i] - a[i];
   for (l = 1; l <= NFix; ++l)
   {
      if (j != l)
      {
         for (i = 1; i <= np; ++i)
            y[i] -= Mufti[ll][l][fixdex[l][i]] * FfDex[ll][l];
      }
   }
}

/**************************************************************************
  
  Modified version of out_gene() installed here 5/26/92
  
  name: out_gene.c
  purpose: subtract major gene effect from the trait.
  date: 10/21/90 (SWG)
  
  Modified  5/26/92 and 6/2/92 (EAT); 
	the program now works on out_fixed() data, including
	subtraction of overall mean.
  
  Modified 10/93, to allow proportional genotypic effects (EAT)
  
***************************************************************************/

void  out_gene (double *ag, double *a, double *mu_vals, int *g, double lam,
                int im)
{
   int   i;

   if (im < 2)
   {
      for (i = 1; i <= np; ++i)
         ag[i] = a[i] - mu_vals[g[idex[i]]];
   }
   else
   {
      for (i = 1; i <= np; ++i)
         ag[i] = a[i] - lam * mu_vals[g[idex[i]]];
   }
}

/*************************************************************************/

double fixeffit (double *z, int j, int l)
/* program for the j th. fixed effect class   for trait l+1 */
{
   double temp;
   int   i, dd, nf[LEVELS + 1];

   temp = 0.0;
   for (i = 1; i <= IFix[j]; ++i)
   {
      nf[i] = 0;
      Mufti[l][j][i] = 0.0;
   }
   for (i = 1; i <= np; ++i)
   {
      dd = fixdex[j][i] * FfDex[l][j];
      ++nf[dd];
      Mufti[l][j][dd] += z[i];
      temp += z[i];
   }

   /* mu[l] = temp/np;  Mufti[l][0][1] = mu[l];  */
   temp = temp / np;

   for (i = 1; i <= IFix[j]; ++i)
   {
      if (nf[i] != 0)
      {
         Mufti[l][j][i] = Mufti[l][j][i] / nf[i];
         Mufti[l][j][i] = Mufti[l][j][i] - temp;
      }
   }
   return (temp);
}
/**************************************************************************/
