/*************************************************************************
  program name: sim_ibd_link.c
  
  Program to do simulate IBD  to estimate probs 
  The state (X) consists of segregation indicators -- 
  X determines the IBD pattern among the specified descendant gametes.
       Up to 8 gametes can be scored for IBD,
          but limit to 6 for easier output.

  Two methods of scoring IBD -- by pattern marginally over loci,
      and by a specified (group of) pattern(s) over windows.

  Original program modified from jntauto, Fall 1999,  EAT.
  New scoring routines added Fall, 1999.
  Combined into single program,  May 2000, EAT.

  ===

  Cleaning up and reformatting the code.
  Unused variables deleted and pointers initialized. 
  Can now be built with MORGAN V2.3 libraries.
                                                  August 2000, N. L.
**************************************************************************/

#define PNAM "sim_ibd_link"
#define DEFINE_DATA

#define DIE {fprintf(stderr, "main prog: calloc failed\nExiting\n"); exit(1);}

#include "msample.h"                    /* MORGAN 2.3 library, NL */

int   main (int argc, char **argv)
{
   int   i, j, k;
   int **dodo = NULL;
   double **quant = NULL;
   int   t, c, nh, pt;
   int   nit;                       /* Number of iterations */
   int   ngenes;                    /* Number of genes */
   int   nfou;                      /* Number of founders */
   int   nloc;                      /* Number of loci */
   int   nsw, *nns;

   double score[NC][NLOC][PT6], sss[NLOC], ss;
   /* ibd counts, each locus, each pattern */
   int   count[NC];
   /* number of steps on each component  */

   double marek[NLOC], rec0[3];
   int   loc0[3];
   int   jj;

   int   npt[NC], list[NC][PT6];    /* list of states to be scored */
   int   npr[NC], probands[NC][G8], posp;  /* indices of inds and gametes */
   int   proband_gamete[NC][G8];    /* to be scored in each component */
   int   index[NC][CD6], label[NC][PT6], gaga[G6];

   int   loc_span;
   FILE *fopen (), *fp;

   print_head (*argv);
   get_seeds ("seed.vary");

   MaxNames = MAX_NAMES;
   NColData = 7;
   MaxNCmps = 10;

   /* input the pedigree, and data observations, */

   if (argc != 2)
   {
      printf ("Usage: sim_ibd_link input_file \n");
   }

   if ((fp = fopen (argv[1], "r")) == NULL)
   {
      fprintf (stderr, "Can't open the data file.\n");
      perror ((char *) NULL);
      exit (1);
   }

   fscanf (fp, "%d %d", &PedSize, &i);  /* i and nt should be 0 */
   fscanf (fp, "%d %d", &NInts, &NTraits);

   /* printf ("\n pedsize =%d, dodo_dimension = %d", PedSize, i); */
   printf ("\n pedsize =%d", PedSize);
   printf ("\n and ints and floats = %d %d \n", NInts, NTraits);
   fflush (stdout);
   for (j = 0; j < NInts; ++j)
   {
      SelInt[j] = j + 1;
   }

   alloc_ped ();
   input_ped (fp, 0, dodo, quant);

   nfou = count_founders ();
   printf ("\n # founders = %d", nfou);
   printf ("\n # components = %d", n_compnts);

   fscanf (fp, "%d %d", &nit, &loc_span);

   fscanf (fp, "%d", &nloc);
   printf ("\n # marker loci = %d", nloc);
   fflush (stdout);

   /* now set and index switches  */

   allocate (pedsize, nloc);

   MaxNsw = 2 * (pedsize - nfou);
   nns = malloc ((n_compnts + 1) * sizeof (int));
   alloc_chk (WHERE, nns);

   nsw = list_switches (nloc, nns);

   printf ("\n have now set switches");
   fflush (stdout);

   /* set up recombination patterns and initial lods */

   for (j = 1; j < nloc; ++j)
   {
      fscanf (fp, "%lf", marek + j);
   }
   marek[nloc] = 0.5;
   nh = 0;
   for (i = 0; i <= nh; ++i)
   {
      fscanf (fp, "%d %lf", loc0 + i, rec0 + i);
   }
   set_chromosomes (nloc, marek, nh, loc0, rec0, 0);

   printf ("\n simulation recombinations:");
   for (j = 0; j < nloc; ++j)
   {
      printf ("  %d %7.4f", chromo[0][j], xecomo[0][0][j]);
   }
   printf ("  %d", chromo[0][nloc]);

   /*   set up probands for gene scoring; one proband per component */
   /*       for scoring over window size loc_span, if loc_span > 1  */
   /*   OR                                                          */
   /*   set up set of proband gametes for scoring                   */
   /*       marginally over loci                                    */

   if (loc_span > 1)
   {
      for (c = 0; c < n_compnts; ++c)
      {
         fscanf (fp, "%d", npr + c);
         printf ("\n %d proband in component %d;", npr[c], c + 1);
         fflush (stdout);
         /* npr[c] is name of proband to be scored in cmpt c */
         posp = posit[npr[c]][0];
         if ((nghd[posp].component) != (c + 1))
         {
            printf ("\n WARNING proband %d not in correct component! %d\n",
                    npr[c], c + 1);
         }
         npr[c] = posp;
      }
   }
   else
   {
      for (c = 0; c < n_compnts; ++c)
      {
         fscanf (fp, "%d", npr + c);
         printf ("\n %d proband gametes in component %d;", npr[c], c + 1);
         fflush (stdout);
         /* npr[c] is number of gametes to be scored in cmpt c */
         for (i = 0; i < npr[c]; ++i)
         {
            fscanf (fp, "%d %d", &probands[c][i],
                    &proband_gamete[c][i]);
            posp = posit[probands[c][i]][0];
            if ((nghd[posp].component) != (c + 1))
            {
               printf ("\n WARNING proband %d not in correct component! %d\n",
                       probands[c][i], c + 1);
            }
            printf ("  %d:%d", probands[c][i], proband_gamete[c][i]);
            probands[c][i] = posp;
         }
      }
   }

   /* realise the initial switches if not already read above */

   realise (nloc, xecomo[0][0], nsw);

   /* set genes from initial switches */

   for (j = 0; j <= nloc; ++j)
   {
      ngenes = set_genes (j);
      printf ("\n \n set genes at locus %d", j);
      fflush (stdout);
   }

   /* probability now computed separately by component  */

   for (c = 0; c < n_compnts; ++c)
   {
      count[c] = 0;
      for (j = 0; j <= nloc; ++j)
      {
         for (pt = 0; pt < PT6; ++pt)
         {
            score[c][j][pt] = 0.0;
         }
      }
   }

   /*  
    *  for (j = 0; j <= nloc; ++j)
    *  {
    *     printf ("\n locus %d set-up \n", j);
    *     print_set_up (j);
    *     printf ("\n");
    *     fflush (stdout);
    *  }
    */

   /*  count total number of patterns -- scoring over loci -- 2**nloc  */

   for (c = 0; c < n_compnts; ++c)
   {
      if (loc_span > 1)
      {
         npt[c] = 1;
         for (j = 0; j < loc_span; ++j)
            npt[c] *= 2;
      }
      else
      {
         /* for JV pedigree npt =15; specified as 4 gametes */
         npt[c] = index_ibd (npr[c], 0, 0, index[c], label[c], list[c], fp);
      }
      printf ("\n  number of scored pattern sets = %d on component %d",
              npt[c], c + 1);
   }

   fflush (stdout);
   /*  close data file and open output file */
   fclose (fp);

   /* Open file for dumping intermediate output */
   /* 
    * if ((fp = fopen (argv[2], "w")) == NULL)
    * {
    *    fprintf (stderr, "Can't open the output file.\n");
    *    perror ((char *) NULL);
    *    exit (1);
    * }
    */

   fflush (stdout);

   printf ("\n now start simulation");
   fflush (stdout);

   /* now test switches and do MC sampling */

   for (t = 0; t < nit; ++t)
   {
      realise (nloc, xecomo[0][0], nsw);

      for (j = 0; j <= nloc; ++j)
      {
         jj = chromo[0][j];
         ngenes = set_genes (jj);
      }

      for (c = 0; c < n_compnts; ++c)
      {
         ++count[c];
      }
      /* score ibd pattern over loci in compt c */
      for (c = 0; c < n_compnts; ++c)
      {
         if (loc_span > 1)
         {
            for (j = (loc_span / 2); j <= (nloc - (loc_span / 2)); ++j)
            {
               pt = pattern (npr[c], j, loc_span);
               ++score[c][j][pt];
            }
         }
         else
         {
            for (j = 0; j <= nloc; ++j)
            {
               jj = chromo[0][j];
               /* scoring for 4C problem on JV pedigree */
               get_genes (npr[c], jj, gaga, probands[c], proband_gamete[c]);
               pt = score_ibd (npr[c], gaga);
               pt = index[c][pt];
               pt = list[c][pt];
               ++score[c][j][pt];
            }
         }
      }
      /* Dumping intermediate output */
      /*
       * if (((t + 1) / fwit) * fwit == (t + 1))
       * {
       *    for (c = 0; c < n_compnts; ++c)
       *    {
       *       for (j = 0; j <= nloc; ++j)
       *       {
       *          jj = chromo[0][j];
       *          fprintf (fp, "\n %d  %d  ", jj, count[c]);
       *          for (pt = 0; pt < npt[c]; ++pt)
       *          {
       *             if (loc_span > 1)
       *                fprintf (fp, " %e", score[c][j][pt]);
       *             else
       *                fprintf (fp, " %e", score[c][jj][list[c][pt]]);
       *          }
       *          fflush (fp);
       *       }
       *    }
       * }
       * fclose (fp);
       */
   }

   for (c = 0; c < n_compnts; ++c)
   {
      printf ("\n  probabilities of ibd patterns of component %d", c + 1);
      printf ("     count = %d", count[c]);
      fflush (stdout);

      if (loc_span > 1)
      {
         for (k = 0; k < npt[c]; ++k)
         {
            binary (k, loc_span, gaga);
            printf ("\n pt = %d, ibd ", k);
            for (j = 0; j < loc_span; ++j)
            {
               printf (" %d", gaga[j]);
            }
            for (j = 1; j < nloc; ++j)
            {
               printf (" %7.4f", score[c][j][k] / count[c]);
            }
            fflush (stdout);
         }
      }
      else
      {
         for (k = 0; k < npt[c]; ++k)
         {
            ss = 0.0;
            pt = list[c][k];
            for (j = 0; j <= nloc; ++j)
            {
               sss[j] = score[c][j][pt] / count[c];
               ss += sss[j];
            }
            if (ss > 0.001)
            {
               invert_label (label[c][k], npr[c], gaga);
               printf ("\n pt %d label %d   ", pt, label[c][k]);
               for (i = 0; i < npr[c]; ++i)
               {
                  printf (" %d", gaga[i]);
               }
               printf (" : ");
               for (j = 0; j <= nloc; ++j)
               {
                  printf ("%7.4f", sss[j]);
               }
               fflush (stdout);
            }
         }
      }
      fflush (stdout);
   }

   /* now we are done  */

   printf ("\n");
   print_tail ();
   put_seeds ("seed.vary");
   fflush (stdout);

   return 0;
}
