/**********************************************************************
 *                               called by MORGAN
 *                               functions or programs.
 *           give_warn()         print warning 
 *           give_warn_where()   messages. 
 *           give_err()          print error
 *           give_err_where()    messages. 
 *           give_up()           these last 2 functions
 *           give_up_where()     also terminate processing
 ********************************************************************** 
 *  arguments are:
 *
 *    give_warn(), give_err(), give_up():
 *        the subscript for the error message in Errrs 
 *        zero, one or more printf arguments 
 *
 *    give_warn_where(), give_err_where(), give_up_where():
 *        the name of the file (__FILE__)
 *        the line no. (__LINE__)
 *        the subsrcipt for the error message in Errrs
 *        zero, one or more printf arguments
 **********************************************************************/

#include "stuff.h"

static void v_prnt_err (int, char);
static void prnt_where (char *, int);
static void prnt_qit (void);
static void flush_line (void);

static va_list pargs;

void  give_warn (int err_no,...)
{
   flush_line ();
   va_start (pargs, err_no);
   v_prnt_err (err_no, 'W');
   va_end (pargs);

   if (++nWarns > MAX_WARNS)
      give_up (ERS_mxw);

   return;
}

void  give_warn_where (char *src_fil, int lin_no, int err_no,...)
{
   flush_line ();
   prnt_where (src_fil, lin_no);

   va_start (pargs, err_no);
   v_prnt_err (err_no, 'W');
   va_end (pargs);

   if (++nWarns > MAX_WARNS)
      give_up (ERS_mxw);

   return;
}

void  give_err (int err_no,...)
{
   flush_line ();
   va_start (pargs, err_no);
   v_prnt_err (err_no, 'E');
   va_end (pargs);

   return;
}

void  give_err_where (char *src_fil, int lin_no, int err_no,...)
{
   flush_line ();
   prnt_where (src_fil, lin_no);

   va_start (pargs, err_no);
   v_prnt_err (err_no, 'E');
   va_end (pargs);

   return;
}

void  give_up (int err_no,...)
{
   flush_line ();
   if (err_no != ERR_qit)
   {
      va_start (pargs, err_no);
      v_prnt_err (err_no, 'F');
      va_end (pargs);
   }
   prnt_qit ();
   exit (EXIT_FAILURE);
}

void  give_up_where (char *src_fil, int lin_no, int err_no,...)
{
   flush_line ();
   prnt_where (src_fil, lin_no);

   va_start (pargs, err_no);
   if (err_no != ERR_qit)
      v_prnt_err (err_no, 'F');
   va_end (pargs);

   prnt_qit ();
   exit (EXIT_FAILURE);
}

void  v_prnt_err (int e_no, char level)
{
   vfprintf (stderr, Errrs[e_no], pargs);
   fprintf (stderr, "  (%c)\n\n", level);
   fflush (stderr);

   return;
}

void  prnt_where (char *src, int lin)
{
   fprintf (stderr, " File %s, line %d:\n", src, lin);
   return;
}

void  prnt_qit (void)
{
   fprintf (stderr, Errrs[ERR_qit]);
   fflush (stderr);

   return;
}

void  flush_line (void)
{
   printf ("\n");
   fflush (stdout);

   return;
}
